""" KNN retrieval using an Annoy index. """

import time
from multiprocessing import Process, cpu_count, Queue
from collections import namedtuple
from operator import attrgetter
from scipy.sparse import issparse
from annoy import AnnoyIndex
from tqdm import tqdm
from tensorflow.keras.utils import HDF5Matrix
import numpy as np


def build_annoy_index(X, path, ntrees=50, build_index_on_disk=True, verbose=1):
    """ Build a standalone annoy index.

    :param array X: numpy array with shape (n_samples, n_features)
    :param str path: The filepath of a trained annoy index file
        saved on disk.
    :param int ntrees: The number of random projections trees built by Annoy to
        approximate KNN. The more trees the higher the memory usage, but the
        better the accuracy of results.
    :param bool build_index_on_disk: Whether to build the annoy index directly
        on disk. Building on disk should allow for bigger datasets to be indexed,
        but may cause issues. If None, on-disk building will be enabled for Linux, 
        but not Windows due to issues on Windows.
    :param int verbose: Controls the volume of logging output the model
        produces when training. When set to 0, silences outputs, when above 0
        will print outputs.

    """

    if not isinstance(X, HDF5Matrix):
        X = X.reshape((X.shape[0], -1))

    index = AnnoyIndex(X.shape[1], metric='angular')
    if build_index_on_disk:
        index.on_disk_build(path)

    if issparse(X):
        for i in tqdm(range(X.shape[0]), disable=verbose < 1):
            vector = X[i].toarray()[0]
            index.add_item(i, vector)
    else:
        for i in tqdm(range(X.shape[0]), disable=verbose < 1):
            vector = X[i]
            index.add_item(i, vector)

    try:
        index.build(ntrees)
    except Exception:
        msg = ("Error building Annoy Index. Passing on_disk_build=False"
               " may solve the issue, especially on Windows.")
        raise IndexBuildingError(msg)
    else:
        if not build_index_on_disk:
            index.save(path)
        return index


def extract_knn(X, index_filepath, k=150, search_k=-1, verbose=1):
    """ Starts multiple processes to retrieve nearest neighbours using
        an Annoy Index in parallel """

    if not isinstance(X, HDF5Matrix):
        X = X.reshape((X.shape[0], -1))

    n_dims = X.shape[1]

    chunk_size = X.shape[0] // cpu_count()
    remainder = (X.shape[0] % cpu_count()) > 0
    process_pool = []
    results_queue = Queue()

    # Split up the indices and assign processes for each chunk
    i = 0
    while (i + chunk_size) <= X.shape[0]:
        process_pool.append(KnnWorker(index_filepath, k, search_k, n_dims,
                                      (i, i+chunk_size), results_queue))
        i += chunk_size
    if remainder:
        process_pool.append(KnnWorker(index_filepath, k, search_k, n_dims,
                                      (i, X.shape[0]), results_queue))

    try:
        for process in process_pool:
            process.start()

        # Read from queue constantly to prevent it from becoming full
        with tqdm(total=X.shape[0], disable=verbose < 1) as pbar:
            neighbour_list = []
            neighbour_list_length = len(neighbour_list)
            while any(process.is_alive() for process in process_pool):
                while not results_queue.empty():
                    neighbour_list.append(results_queue.get())
                progress = len(neighbour_list) - neighbour_list_length
                pbar.update(progress)
                neighbour_list_length = len(neighbour_list)
                time.sleep(0.1)

            while not results_queue.empty():
                neighbour_list.append(results_queue.get())

        neighbour_list = sorted(neighbour_list, key=attrgetter('row_index'))
        neighbour_list = list(map(attrgetter('neighbour_list'), neighbour_list))

        return np.array(neighbour_list)

    except:
        print('Halting KNN retrieval and cleaning up')
        for process in process_pool:
            process.terminate()
        raise


IndexNeighbours = namedtuple('IndexNeighbours', 'row_index neighbour_list')


class KnnWorker(Process):
    """
    Upon construction, this worker process loads an annoy index from disk.
    When started, the neighbours of the data-points specified by `data_indices`
    will be retrieved from the index according to the provided parameters
    and stored in the 'results_queue'.

    `data_indices` is a tuple of integers denoting the start and end range of
    indices to retrieve.
    """
    def __init__(self, index_filepath, k, search_k, n_dims,
                 data_indices, results_queue):
        self.index_filepath = index_filepath
        self.k = k
        self.n_dims = n_dims
        self.search_k = search_k
        self.data_indices = data_indices
        self.results_queue = results_queue
        super(KnnWorker, self).__init__()

    def run(self):
        try:
            index = AnnoyIndex(self.n_dims, metric='angular')
            index.load(self.index_filepath)
            for i in range(self.data_indices[0], self.data_indices[1]):
                neighbour_indexes = index.get_nns_by_item(
                    i, self.k, search_k=self.search_k, include_distances=False)
                neighbour_indexes = np.array(neighbour_indexes,
                                             dtype=np.uint32)
                self.results_queue.put(
                    IndexNeighbours(row_index=i,
                                    neighbour_list=neighbour_indexes))
        except Exception as e:
            self.exception = e
        finally:
            self.results_queue.close()


class IndexBuildingError(OSError):
    pass



import pandas as pd
import numpy as np

try:
    import param
except ImportError:
    raise ImportError(
        'Failed to import param. You must run '
        '`pip install param` to use vis_utils')

try:
    import panel as pn
except ImportError:
    pn = None

try:
    import holoviews as hv
    from holoviews.operation.datashader import datashade, dynspread
    hv.extension('bokeh')
except ImportError:
    hv = None

try:
    import datashader as ds
    from datashader.colors import Sets1to3
except ImportError:
    ds = None

def check_dependencies():
    if hv is None:
        return False
    if ds is None:
        return False
    if param is None:
        return False
    if pn is None:
        return False
    return True


class EmbeddingsExplorer(param.Parameterized):
    label_flag = param.Boolean(default=True)

    def __init__(self):
        super(EmbeddingsExplorer, self).__init__()
        if not check_dependencies():
            message = (
                'Failed to import dependencies. You must '
                '`pip install holoviews datashader bokeh` '
                'for `EmbeddingsExplorer` to work.')
            raise ImportError(message)
        self.embeddings = None
        self.classes = None

    @param.depends('label_flag')
    def get_points(self):
        embeddings = self.embeddings
        classes = self.classes
        if (self.label_flag) and (classes is not None):
            data = pd.DataFrame(embeddings)
            data.columns = ['ivis 1', 'ivis 2']
            data['label'] = classes
            num_ks = len(np.unique(classes))
            color_key = list(enumerate(Sets1to3[0:num_ks]))

            embed = {k: hv.Points(data.values[classes == k, :], ['ivis 1', 'ivis 2'], 'k',
                                  label=str(k)).opts(color=v, size=0) for k, v in color_key}
            dse = dynspread(
                datashade(hv.NdOverlay(embed, kdims=['k']), aggregator=ds.by('k', ds.count())))
            color_points = hv.NdOverlay(
                {k: hv.Points([0, 0]).opts(color=v, size=0) for k, v in color_key})
            points = color_points * dse
        else:
            points = datashade(hv.Points(embeddings))

        points.opts(height=400, width=500, xaxis=None, yaxis=None)
        return points

    def view(self, embeddings, classes=None):
        self.embeddings = embeddings
        self.classes = classes
        return pn.Row(self.get_points, self.param)



"""
Triplet generators.

Functions for creating generators that will yield batches of triplets.

Triplets will be created using KNNs, which can either be precomputed or
dynamically generated.
- generate_knn_triplets_from_neighbour_list will precompute KNNs
- generate_knn_triplets_from_annoy_index will dynamically generate KNNs

Where possible, precomputed KNNs are advised for speed, but where memory is
a concern, dynamically generated triplets can be useful.

"""

import numpy as np
import tensorflow as tf
from annoy import AnnoyIndex
from tensorflow.keras.utils import Sequence
from scipy.sparse import issparse
from .knn import extract_knn


def generator_from_knn_matrix(X, Y, neighbour_matrix, k, batch_size):
    if k >= X.shape[0] - 1:
        raise Exception('''k value greater than or equal to (num_rows - 1)
                        (k={}, rows={}). Lower k to a smaller
                        value.'''.format(k, X.shape[0]))
    if batch_size > X.shape[0]:
        raise Exception('''batch_size value larger than num_rows in dataset
                        (batch_size={}, rows={}). Lower batch_size to a
                        smaller value.'''.format(batch_size, X.shape[0]))
    if Y is None:
        return create_knn_triplet_dataset(X, neighbour_matrix, batch_size=batch_size)

    return create_labeled_knn_triplet_dataset(X, Y, neighbour_matrix,
                                              batch_size=batch_size)


def generator_from_index(X, Y, index_path, k, batch_size, search_k=-1,
                         precompute=True, verbose=1):
    if k >= X.shape[0] - 1:
        raise Exception('''k value greater than or equal to (num_rows - 1)
                        (k={}, rows={}). Lower k to a smaller
                        value.'''.format(k, X.shape[0]))
    if batch_size > X.shape[0]:
        raise Exception('''batch_size value larger than num_rows in dataset
                        (batch_size={}, rows={}). Lower batch_size to a
                        smaller value.'''.format(batch_size, X.shape[0]))

    if Y is None:
        if precompute:
            if verbose > 0:
                print('Extracting KNN from index')

            neighbour_matrix = extract_knn(X, index_path, k=k,
                                           search_k=search_k, verbose=verbose)
            return create_knn_triplet_dataset(X, neighbour_matrix, batch_size=batch_size)

        index = AnnoyIndex(X.shape[1], metric='angular')
        index.load(index_path)
        return create_annoy_triplet_dataset(X, index, k=k,
                                            batch_size=batch_size,
                                            search_k=search_k)
    else:
        if precompute:
            if verbose > 0:
                print('Extracting KNN from index')

            neighbour_matrix = extract_knn(X, index_path, k=k,
                                           search_k=search_k, verbose=verbose)
            return create_labeled_knn_triplet_dataset(X, Y, neighbour_matrix,
                                                      batch_size=batch_size)

        index = AnnoyIndex(X.shape[1], metric='angular')
        index.load(index_path)
        return create_labeled_annoy_triplet_dataset(X, Y, index,
                                                    k=k, batch_size=batch_size,
                                                    search_k=search_k)

def create_annoy_triplet_dataset(X, annoy_index, k=150, batch_size=32, search_k=-1):
    knn_sequence = AnnoyTripletGenerator(
        X, annoy_index, k=k, batch_size=batch_size, search_k=search_k)

    def get_triplets_by_index(index):
        triplets, labels = knn_sequence[index]
        return (*triplets, labels)

    def tf_get_triplets_by_index(index):
        anchors, positives, negatives, labels = tf.py_function(
            get_triplets_by_index, [index],
            [tf.float32, tf.float32, tf.float32, tf.float32],
        )
        anchors.set_shape([None, *X.shape[1:]])
        positives.set_shape([None, *X.shape[1:]])
        negatives.set_shape([None, *X.shape[1:]])
        labels.set_shape([None,])
        return tuple([anchors, positives, negatives]), labels

    index_dataset = tf.data.Dataset.from_tensor_slices(
        list(range(int(np.ceil(len(X) / batch_size)))))
    index_dataset = index_dataset.shuffle(int(np.ceil(len(X) / batch_size)))
    index_dataset = index_dataset.repeat()

    dataset = index_dataset.map(
        tf_get_triplets_by_index,
        num_parallel_calls=tf.data.experimental.AUTOTUNE)
    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
    return dataset

class AnnoyTripletGenerator(Sequence):

    def __init__(self, X, annoy_index, k=150, batch_size=32, search_k=-1):
        self.X = X
        self.annoy_index = annoy_index
        self.k = k
        self.batch_size = batch_size
        self.search_k = search_k
        self.placeholder_labels = np.empty(batch_size, dtype=np.uint8)

    def __len__(self):
        return int(np.ceil(self.X.shape[0] / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_indices = range(idx * self.batch_size,
                              min((idx + 1) * self.batch_size, self.X.shape[0]))

        placeholder_labels = self.placeholder_labels[:len(batch_indices)]
        triplet_batch = [self.knn_triplet_from_annoy_index(row_index)
                         for row_index in batch_indices]

        if issparse(self.X):
            triplet_batch = [[e.toarray()[0] for e in t] for t in triplet_batch]

        triplet_batch = np.array(triplet_batch)

        return tuple([triplet_batch[:, 0], triplet_batch[:, 1], triplet_batch[:, 2]]), placeholder_labels

    def knn_triplet_from_annoy_index(self, row_index):
        """ A random (unweighted) positive example chosen. """
        triplet = []
        neighbour_list = self.annoy_index.get_nns_by_item(
            row_index, self.k + 1, search_k=self.search_k, include_distances=False)
        neighbour_list = np.array(neighbour_list, dtype=np.uint32)

        # Take a random neighbour as positive
        neighbour_ind = np.random.choice(neighbour_list)

        # Take a random non-neighbour as negative
        # Pick a random index until one fits constraint, usually faster.
        negative_ind = np.random.randint(0, self.X.shape[0])
        while negative_ind in neighbour_list:
            negative_ind = np.random.randint(0, self.X.shape[0])

        triplet += [self.X[row_index], self.X[neighbour_ind], self.X[negative_ind]]
        return triplet


def create_knn_triplet_dataset(X, neighbour_matrix, batch_size=32):
    knn_sequence = KnnTripletGenerator(X, neighbour_matrix, batch_size=batch_size)

    def get_triplets_by_index(index):
        triplets, labels = knn_sequence[index]
        return (*triplets, labels)

    def tf_get_triplets_by_index(index):
        anchors, positives, negatives, labels = tf.py_function(
            get_triplets_by_index, [index],
            [tf.float32, tf.float32, tf.float32, tf.float32],
        )
        anchors.set_shape([None, *X.shape[1:]])
        positives.set_shape([None, *X.shape[1:]])
        negatives.set_shape([None, *X.shape[1:]])
        labels.set_shape([None,])
        return tuple([anchors, positives, negatives]), labels

    index_dataset = tf.data.Dataset.from_tensor_slices(
        list(range(int(np.ceil(len(X) / batch_size)))))
    index_dataset = index_dataset.shuffle(int(np.ceil(len(X) / batch_size)))
    index_dataset = index_dataset.repeat()

    dataset = index_dataset.map(
        tf_get_triplets_by_index,
        num_parallel_calls=tf.data.experimental.AUTOTUNE)
    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
    return dataset


class KnnTripletGenerator(Sequence):

    def __init__(self, X, neighbour_matrix, batch_size=32):
        self.X = X
        self.neighbour_matrix = neighbour_matrix
        self.batch_size = batch_size
        self.placeholder_labels = np.empty(batch_size, dtype=np.uint8)

    def __len__(self):
        return int(np.ceil(self.X.shape[0] / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_indices = range(idx * self.batch_size, min((idx + 1) * self.batch_size, self.X.shape[0]))

        placeholder_labels = self.placeholder_labels[:len(batch_indices)]
        triplet_batch = [self.knn_triplet_from_neighbour_list(row_index, self.neighbour_matrix[row_index])
                         for row_index in batch_indices]

        if issparse(self.X):
            triplet_batch = [[e.toarray()[0] for e in t] for t in triplet_batch]
        triplet_batch = np.array(triplet_batch)

        return tuple([triplet_batch[:, 0], triplet_batch[:, 1], triplet_batch[:, 2]]), placeholder_labels

    def knn_triplet_from_neighbour_list(self, row_index, neighbour_list):
        """ A random (unweighted) positive example chosen. """
        triplets = []

        # Take a random neighbour as positive
        neighbour_ind = np.random.choice(neighbour_list)

        # Take a random non-neighbour as negative
        # Pick a random index until one fits constraint. An optimization.
        negative_ind = np.random.randint(0, self.X.shape[0])
        while negative_ind in neighbour_list:
            negative_ind = np.random.randint(0, self.X.shape[0])

        triplets += [self.X[row_index], self.X[neighbour_ind],
                     self.X[negative_ind]]
        return triplets


def create_labeled_annoy_triplet_dataset(X, Y, annoy_index, k=150, batch_size=32, search_k=-1):
    knn_sequence = LabeledAnnoyTripletGenerator(
        X, Y, annoy_index, k=k, batch_size=batch_size, search_k=search_k)

    def get_triplets_by_index(index):
        triplets, labels = knn_sequence[index]
        return (*triplets, labels[0])

    def tf_get_triplets_by_index(index):
        anchors, positives, negatives, labels = tf.py_function(
            get_triplets_by_index, [index],
            [tf.float32, tf.float32, tf.float32, tf.float32],
        )
        anchors.set_shape([None, *X.shape[1:]])
        positives.set_shape([None, *X.shape[1:]])
        negatives.set_shape([None, *X.shape[1:]])
        if Y.ndim > 1:
            labels.set_shape([None, Y.shape[-1]])
        else:
            labels.set_shape([None,])
        return tuple([anchors, positives, negatives]), tuple([labels, labels])

    index_dataset = tf.data.Dataset.from_tensor_slices(
        list(range(int(np.ceil(len(X) / batch_size)))))
    index_dataset = index_dataset.shuffle(int(np.ceil(len(X) / batch_size)))
    index_dataset = index_dataset.repeat()

    dataset = index_dataset.map(
        tf_get_triplets_by_index,
        num_parallel_calls=tf.data.experimental.AUTOTUNE)
    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
    return dataset


class LabeledAnnoyTripletGenerator(Sequence):

    def __init__(self, X, Y, annoy_index, k=150, batch_size=32, search_k=-1):
        self.X, self.Y = X, Y
        self.annoy_index = annoy_index
        self.k = k
        self.batch_size = batch_size
        self.search_k = search_k

    def __len__(self):
        return int(np.ceil(self.X.shape[0] / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_indices = range(idx * self.batch_size,
                              min((idx + 1) * self.batch_size,
                                  self.X.shape[0]))

        label_batch = self.Y[batch_indices]
        triplet_batch = [self.knn_triplet_from_annoy_index(row_index)
                         for row_index in batch_indices]

        if issparse(self.X):
            triplet_batch = [[e.toarray()[0] for e in t] for t in triplet_batch]

        triplet_batch = np.array(triplet_batch)

        return tuple([triplet_batch[:, 0], triplet_batch[:, 1], triplet_batch[:, 2]]), tuple([np.array(label_batch), np.array(label_batch)])

    def knn_triplet_from_annoy_index(self, row_index):
        """ A random (unweighted) positive example chosen. """
        triplet = []
        neighbour_list = self.annoy_index.get_nns_by_item(
            row_index, self.k + 1, search_k=self.search_k, include_distances=False)
        neighbour_list = np.array(neighbour_list, dtype=np.uint32)

        # Take a random neighbour as positive
        neighbour_ind = np.random.choice(neighbour_list)

        # Take a random non-neighbour as negative
        # Pick a random index until one fits constraint, usually faster
        negative_ind = np.random.randint(0, self.X.shape[0])
        while negative_ind in neighbour_list:
            negative_ind = np.random.randint(0, self.X.shape[0])

        triplet += [self.X[row_index], self.X[neighbour_ind], self.X[negative_ind]]
        return triplet


def create_labeled_knn_triplet_dataset(X, Y, neighbour_matrix, batch_size=32):
    knn_sequence = LabeledKnnTripletGenerator(X, Y, neighbour_matrix, batch_size=batch_size)

    def get_triplets_by_index(index):
        triplets, labels = knn_sequence[index]
        return (*triplets, labels[0])

    def tf_get_triplets_by_index(index):
        anchors, positives, negatives, labels = tf.py_function(
            get_triplets_by_index, [index],
            [tf.float32, tf.float32, tf.float32, tf.float32],
        )
        anchors.set_shape([None, *X.shape[1:]])
        positives.set_shape([None, *X.shape[1:]])
        negatives.set_shape([None, *X.shape[1:]])
        if Y.ndim > 1:
            labels.set_shape([None, Y.shape[-1]])
        else:
            labels.set_shape([None,])
        return tuple([anchors, positives, negatives]), tuple([labels, labels])

    index_dataset = tf.data.Dataset.from_tensor_slices(
        list(range(int(np.ceil(len(X) / batch_size)))))
    index_dataset = index_dataset.shuffle(int(np.ceil(len(X) / batch_size)))
    index_dataset = index_dataset.repeat()

    dataset = index_dataset.map(
        tf_get_triplets_by_index,
        num_parallel_calls=tf.data.experimental.AUTOTUNE)
    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
    return dataset


class LabeledKnnTripletGenerator(Sequence):

    def __init__(self, X, Y, neighbour_matrix, batch_size=32):
        self.X, self.Y = X, Y
        self.neighbour_matrix = neighbour_matrix
        self.batch_size = batch_size

    def __len__(self):
        return int(np.ceil(self.X.shape[0] / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_indices = range(idx * self.batch_size, min((idx + 1) * self.batch_size, self.X.shape[0]))

        label_batch = self.Y[batch_indices]
        triplet_batch = [self.knn_triplet_from_neighbour_list(row_index, self.neighbour_matrix[row_index])
                         for row_index in batch_indices]

        if issparse(self.X):
            triplet_batch = [[e.toarray()[0] for e in t] for t in triplet_batch]

        triplet_batch = np.array(triplet_batch)

        return tuple([triplet_batch[:, 0], triplet_batch[:, 1], triplet_batch[:, 2]]), tuple([np.array(label_batch), np.array(label_batch)])

    def knn_triplet_from_neighbour_list(self, row_index, neighbour_list):
        """ A random (unweighted) positive example chosen. """
        triplets = []

        # Take a random neighbour as positive
        neighbour_ind = np.random.choice(neighbour_list)

        # Take a random non-neighbour as negative
        # Pick a random index until one fits constraint. An optimization.
        negative_ind = np.random.randint(0, self.X.shape[0])
        while negative_ind in neighbour_list:
            negative_ind = np.random.randint(0, self.X.shape[0])

        triplets += [self.X[row_index],
                     self.X[neighbour_ind],
                     self.X[negative_ind]]
        return triplets



"""Contains helper functions for creating Siamese networks from base network architectures.
Additionally, provides several base network constructors as default options."""

from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense, AlphaDropout, Lambda
from tensorflow.keras import backend as K

from tensorflow.keras.regularizers import l2


def triplet_network(base_network, embedding_dims=2, embedding_l2=0.0):
    """ Creates a triplet Siamese Neural Network from a base_network.
    The base network will have an extra Dense layer of the requested embedding_dims added to
    the end if embedding_dims is not None.

    The outputs of the three network heads will be stacked into the shape:
    (3, batch_size, embedding_dims) unless embedding_dims is None (in which case the existing dims
    of last base_network layer will be used).

    Outputs: tuple(
        model: tf.keras.models.Model. The constructed triplet Siamese network
        processed_a: tf.keras.layers.Dense. Result of applying the base_network to anchor input.
        processed_p: tf.keras.layers.Dense. Result of applying the base_network to positive input.
        processed_n: tf.keras.layers.Dense. Result of applying the base_network to negative input.
    ) """

    def output_shape(shapes):
        shape1, _, _ = shapes
        return (3, shape1[0],)

    input_a = Input(shape=base_network.input_shape[1:])
    input_p = Input(shape=base_network.input_shape[1:])
    input_n = Input(shape=base_network.input_shape[1:])

    if embedding_dims is None:
        embeddings = base_network.output
    else:
        embeddings = Dense(embedding_dims,
                           kernel_regularizer=l2(embedding_l2))(base_network.output)

    network = Model(base_network.input, embeddings)

    processed_a = network(input_a)
    processed_p = network(input_p)
    processed_n = network(input_n)

    triplet = Lambda(K.stack,
                     output_shape=output_shape,
                     name='stacked_triplets')([processed_a,
                                               processed_p,
                                               processed_n],)
    model = Model([input_a, input_p, input_n], triplet)

    return model, processed_a, processed_p, processed_n


def base_network(model_name, input_shape):
    '''Return the defined base_network defined by the model_name string.
    '''
    # Select network to return in if statements to avoid running and
    # constructing unnecessary models as would occur in a dict.
    if model_name == 'szubert':
        return szubert_base_network(input_shape)
    if model_name == 'hinton':
        return hinton_base_network(input_shape)
    if model_name == 'maaten':
        return maaten_base_network(input_shape)

    raise NotImplementedError(
        'Base network {} is not implemented'.format(model_name))


def get_base_networks():
    return ['szubert', 'hinton', 'maaten']


def szubert_base_network(input_shape):
    '''A small, quick-to-train base network. The default for Ivis.'''
    inputs = Input(shape=input_shape)
    x = Dense(128, activation='selu',
              kernel_initializer='lecun_normal')(inputs)
    x = AlphaDropout(0.1)(x)
    x = Dense(128, activation='selu',
              kernel_initializer='lecun_normal')(x)
    x = AlphaDropout(0.1)(x)
    x = Dense(128, activation='selu', kernel_initializer='lecun_normal')(x)
    return Model(inputs, x)


def hinton_base_network(input_shape):
    '''A base network inspired by the autoencoder architecture published in Hinton's paper
    'Reducing Dimensionality of Data with Neural Networks' (https://www.cs.toronto.edu/~hinton/science.pdf)'''
    inputs = Input(shape=input_shape)
    x = Dense(2000, activation='selu',
              kernel_initializer='lecun_normal')(inputs)
    x = AlphaDropout(0.1)(x)
    x = Dense(1000, activation='selu', kernel_initializer='lecun_normal')(x)
    x = AlphaDropout(0.1)(x)
    x = Dense(500, activation='selu', kernel_initializer='lecun_normal')(x)
    return Model(inputs, x)


def maaten_base_network(input_shape):
    '''A base network inspired by the network architecture published in Maaten's t-SNE paper
    'Learning a Parametric Embedding by Preserving Local Structure'
    (https://lvdmaaten.github.io/publications/papers/AISTATS_2009.pdf)'''
    inputs = Input(shape=input_shape)
    x = Dense(500, activation='selu',
              kernel_initializer='lecun_normal')(inputs)
    x = AlphaDropout(0.1)(x)
    x = Dense(500, activation='selu', kernel_initializer='lecun_normal')(x)
    x = AlphaDropout(0.1)(x)
    x = Dense(2000, activation='selu', kernel_initializer='lecun_normal')(x)
    return Model(inputs, x)
